use structopt::StructOpt;
use std::{
    path::PathBuf,
    boxed::Box,
    error::Error,
};
use serde::{
    Deserialize,
    Serialize,
};
use sqlx::{
    prelude::*,
    sqlite::{
        SqliteConnection,
        SqliteConnectOptions,
        SqliteQueryResult,
    },
};
use tokio::{
    sync::RwLock,
    fs::File,
    io::AsyncReadExt,
};
use crate::cli::Options;

mod cli;

const DEFAULT_DATABASE_PATH: &'static str = "/var/lib/noise-server/noise-server.sqlite";

#[derive(Debug, Serialize, Deserialize)]
#[serde(default, rename_all = "kebab-case")]
struct Config {
    database_path: PathBuf,
}

impl Default for Config {
    fn default() -> Self {
        Config {
            database_path: DEFAULT_DATABASE_PATH.into(),
        }
    }
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn Error>> {
    let clap_matches = Options::clap().get_matches();
    let options = Options::from_clap(&clap_matches);

    if options.generate_config {
        println!("{}", toml::to_string_pretty(&Config::default())?);
        return Ok(());
    }

    let mut config = get_config(&options).await?;

    if clap_matches.is_present("database-path") {
        config.database_path = options.database_path;
    }

    #[cfg(debug_assertions)]
    println!("{:?}", config);

    #[cfg(feature = "sqlite")]
    let mut sqlite_connection = RwLock::new(connect_sqlite(&config.database_path).await?);

    #[cfg(feature = "sqlite")]
    load_sqlite_template(sqlite_connection.get_mut()).await?;

    Ok(())
}

/// Prepare sqlite database connection
#[cfg(feature = "sqlite")]
async fn connect_sqlite(path: &PathBuf) -> sqlx::Result<SqliteConnection> {
    let connect_options = SqliteConnectOptions::new()
        .filename(&path)
        .create_if_missing(true);
    SqliteConnection::connect_with(&connect_options).await
}

#[cfg(feature = "sqlite")]
async fn load_sqlite_template(sqlite_handle: &mut SqliteConnection) -> sqlx::Result<SqliteQueryResult> {
    sqlx::query(concat!(
        include_str!("sqlite/users.sql"),
        include_str!("sqlite/security_preferences.sql"),
        include_str!("sqlite/privacy_preferences.sql"),
        include_str!("sqlite/notification_preferences.sql"),
        include_str!("sqlite/external_servers_privacy_preferences.sql"),
        include_str!("sqlite/external_servers_preferences.sql"),
    )).execute(sqlite_handle).await
}

async fn get_config(options: &Options) -> std::io::Result<Config> {
    if options.no_config {
        Ok(Config::default())
    } else {
        let mut config_file = File::open(&options.config).await?;
        let mut config_string = String::new();
        config_file.read_to_string(&mut config_string).await?;

        toml::from_str(&config_string)
            .map_err(|err| std::io::Error::new(
                std::io::ErrorKind::InvalidData,
                err
            )
        )
    }
}